#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoutil.symbols import Unset

from odoo.release import version_info as ODOO_VERSION_INFO
from odoo.fields import Field as Base


class PropertyField(Base):
    """A property-like field.

    This is always non-store field.  In fact, you may only pass the getter,
    setter and deleter functions.

    .. warning:: This is a kind of hack.

       Normal Python `properties <property>`:class: are not copied through
       Odoo inheritance mechanisms.  This `Property` is a Field and Odoo
       accounts for it.  But, it is NOT an actual ORM field in the sense that
       it never interacts with the DB.  Therefore:

       - You MUST NEVER use ``write()`` or ``create()`` on these type of
         fields.  You may update the value via the `setter` (i.e assign it).

       - You cannot ``search()`` for this type of fields.

       - You cannot use Property fields in views.

         The main difference to any other computed field is that Property is
         untyped.  We don't enforce any type check on the value returned by
         getters.  But, then, we cannot communicate which widget can manage
         the value of a Property.

    The `getter`, `setter` and `deleter` functions receive (and thus we
    required it) a singleton recordset instance.  The `onsetup` function
    receive the field instance, and the model on which the field is being
    setup.

    Usage::

       @Property
       def result(self):
           return 1

       @result.setter
       def result(self, value):
          pass

       @Property(memoize=True)
       def sentinel(self):
           return object()

    You may also do::

       def _set_result(self, value):
           pass

       result = Property(lambda s: 1, setter=_set_result)
       del _set_result

    :keyword memoize: If True, the property will cache the result.  The
             default is False (compute each time the property is read).

    .. versionchanged:: 0.58.0 Added keyword parameter `memoize`.

    .. versionchanged:: 0.67.0 Invalidate Odoo's cache of dependant computable
       fields.

    .. versionchanged:: 0.68.0 Actively trigger recomputation on dependant
       fields.

    """

    # This is the best of the three major versions of Odoo we support.  This
    # create a __slot__ and avoids that these values go to the
    # __getattr__/__setattr__ mechanism in fields.Field.  But it also, means
    # that we need to do some tricks in the `__init__`:meth: to make all
    # versions happy.
    _slots = {
        "property_getter": None,
        "property_setter": None,
        "property_deleter": None,
        "property_onsetup": None,
        "memoize_result": False,
    }
    type = "python-property"  # needed to satisfy ir.models.field

    def __init__(
        self, getter, setter=None, deleter=None, onsetup=None, memoize=False, **kwargs
    ):
        # Notice we don't abide by the expected fields signature.  Instead, we
        # require one that is compatible with `property`; but we ensure that
        # Odoo sees this Property as normal field with custom attributes.  We
        # allow for inheritance (_inherits) of this field.  Such inheritance
        # is done via a related field, but otherwise custom `write/create`
        # that treat properties could be ignored.
        kw = {
            "inherited": kwargs.get("inherited", None),
            "related": kwargs.get("related", None),
            "related_sudo": kwargs.get("related_sudo", False),
        }
        super(PropertyField, self).__init__(
            # Odoo ignores arguments which are None, therefore, let's force
            # them with Unset.  Odoo 9 and 10, uses this args to call
            # field.new(**args), and also to restore the values of the _slots
            # to provided values instead of default ones.  This last thing is
            # done when building the specific field for each Model Class (see
            # our `registry.rst`); cf. Field._setup_attrs.
            #
            # WARNING: If you remove this thinking that put those values just
            # a few lines below, you're fooling yourself.
            property_getter=getter or Unset,
            property_setter=setter or Unset,
            property_deleter=deleter or Unset,
            property_onsetup=onsetup or Unset,
            memoize_result=memoize,
            compute=getter,
            store=False,
            copy=False,
            **kw,
        )
        # In any case, we need those values now to make `new()` below work.
        self.property_getter = getter or Unset
        self.property_setter = setter or Unset
        self.property_deleter = deleter or Unset
        self.property_onsetup = onsetup or Unset
        self.memoize_result = memoize

    def new(self, **kwargs):
        # Ensure the property getter, setter and deleter are provided.  This
        # is used in `setter`:meth: and `deleter`:meth:.
        getter = kwargs.pop("getter", self.property_getter)
        setter = kwargs.pop("setter", self.property_setter)
        deleter = kwargs.pop("deleter", self.property_deleter)
        onsetup = kwargs.pop("onsetup", self.property_onsetup)
        memoize = kwargs.pop("memoize", self.memoize_result)
        return type(self)(
            getter=getter,
            setter=setter,
            deleter=deleter,
            onsetup=onsetup,
            memoize=memoize,
            **kwargs,
        )

    def setup_full(self, model):
        res = super(PropertyField, self).setup_full(model)
        if self.property_onsetup:
            self.property_onsetup(self, model)
        return res

    def setter(self, f):
        return self.new(setter=f)

    def deleter(self, f):
        return self.new(deleter=f)

    def onsetup(self, f):
        return self.new(onsetup=f)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            if not instance:
                return self.null(instance.env)
            instance.ensure_one()
            if not self.memoize_result:
                return self.property_getter(instance)
            else:
                Unset = object()
                result = _get_from_cache(instance, self, Unset)
                if result is Unset:
                    result = self.property_getter(instance)
                    _set_to_cache(instance, self, result)
                return result

    def __set__(self, instance, value):
        if self.property_setter:
            instance.ensure_one()
            self.property_setter(instance, value)
            # In normal Odoo fields, __set__ would call `write` to set the
            # value to the DB, and it is there where recomputation happens.
            # But we don't call write because Property is never stored in the
            # DB, so we need to trigger recomputations ourselves.
            self._cache_and_recompute_dependants(instance, value)
        else:
            raise TypeError("Setting to read-only Property")

    def __delete__(self, instance):
        if self.property_deleter:
            instance.ensure_one()
            self.property_deleter(instance)
            self._cache_and_recompute_dependants(instance, Removed)
        else:
            raise TypeError("Deleting undeletable Property")

    if ODOO_VERSION_INFO < (11, 0):

        def _cache_and_recompute_dependants(self, instance, value):
            spec = [
                (field, ids)
                for field, ids in self.modified(instance)
                # The result of `modified` may include Property fields with
                # memoize_result set; we don't to invalidate those.
                if not isinstance(field, PropertyField) or not field.memoize_result
            ]
            _invalidate_cache(instance, spec)
            if self.memoize_result:
                if value is Removed:
                    _del_from_cache(instance, self)
                else:
                    _set_to_cache(instance, self, value)
            if instance.env.recompute and instance._context.get("recompute", True):
                instance.recompute()

    else:
        # Since Odoo 11 (commit e25bcf41e2f312536a00ef8531b2f6c6e0334ebc),
        # fields no longer have a `modified()` method.  Instead they merged
        # that into the `modified()` method of models; but
        # `BaseModel.modified()` calls the invalidation of the cache without
        # regards of our semantics for memoization.
        #
        # So we need to reset the cache before triggering the recomputation of
        # dependant fields.
        def _cache_and_recompute_dependants(self, instance, value):
            instance.modified([self.name])
            if self.memoize_result:
                if value is not Removed:
                    _set_to_cache(instance, self, value)
                # we don't need to call _del_from_cache because
                # `instance.modified` did it already.
            if instance.env.recompute and instance._context.get("recompute", True):
                instance.recompute()


class _PropertyType(type):
    def __instancecheck__(self, instance):
        return isinstance(instance, PropertyField)

    def __subclasscheck__(self, klass):
        return issubclass(klass, PropertyField)

    def __call__(self, getter=None, setter=None, deleter=None, onsetup=None, **kwargs):
        memoize = kwargs.pop("memoize", Unset)
        if memoize is Unset and getter is None:
            raise TypeError("getter must be provided")
        elif getter is None:

            def result(getter):
                return PropertyField(
                    getter,
                    setter=setter,
                    deleter=deleter,
                    onsetup=onsetup,
                    memoize=bool(memoize),
                    **kwargs,
                )

            return result
        else:
            return PropertyField(
                getter,
                setter=setter,
                deleter=deleter,
                onsetup=onsetup,
                memoize=bool(memoize),
                **kwargs,
            )


class Property(metaclass=_PropertyType):
    # So that isinstance(x, Property) work
    pass


Property.__doc__ = PropertyField.__doc__


if ODOO_VERSION_INFO < (11, 0):

    def _get_from_cache(record, field, default):
        return record.env.cache[field].get(record.id, default)

    def _set_to_cache(record, field, value):
        record.env.cache[field][record.id] = value

    def _del_from_cache(record, field):
        del record.env.cache[field][record.id]

    def _invalidate_cache(record, spec):
        record.env.invalidate(spec)


else:

    def _get_from_cache(record, field, default):
        return record.env.cache.get_value(record, field, default)

    def _set_to_cache(record, field, value):
        record.env.cache.set(record, field, value)

    def _del_from_cache(record, field):
        record.env.cache.remove(record, field)

    def _invalidate_cache(record, spec):
        record.env.cache.invalidate(spec)


# Sentinel object to know we're removing a key from the cache
Removed = object()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from hypothesis import given, strategies as st, assume

from xoeuf.odoo.tests.common import TransactionCase
from xoeuf.odoo.addons.test_property.models import Undefined


values = st.floats(allow_infinity=False, allow_nan=False) | st.integers()


class TestPropertyDepends(TransactionCase):
    @given(values, values)
    def test_setting_property(self, value1, value2):
        assume(value1 != value2)
        Value = self.env["test.property.value"]
        obj = Value.create({})

        obj.price = value1
        self.assertEqual(obj.price_display, "$ {:.2f}".format(value1))

        # The test above won't actually fail because when doing obj.price the
        # value won't be in the cache, so Odoo will trigger the computation,
        # but if we change the value again, the Property must trigger the
        # recomputation itself:
        obj.price = value2
        self.assertEqual(obj.price_display, "$ {:.2f}".format(value2))

        obj.price = Undefined
        self.assertEqual(obj.price_display, "--")

    @given(values, values)
    def test_setting_dependant(self, value1, value2):
        Value = self.env["test.property.value"]
        obj = Value.create({})
        obj.price_display = "alsdkjfalsd"
        self.assertIs(obj.price, Undefined)

    @given(values, values)
    def test_setting_property_and_reading(self, value1, value2):
        assume(value1 != value2)
        Value = self.env["test.property.value"]
        obj = Value.create({})

        obj.price = value1
        self.assertEqual(obj.price_display, "$ {:.2f}".format(value1))

        # The test above won't actually fail because when doing obj.price the
        # value won't be in the cache, so Odoo will trigger the computation,
        # but if we change the value again, the Property must trigger the
        # recomputation itself:
        obj.price = value2
        price_display_stored = obj.read(("price_display_stored",))[0][
            "price_display_stored"
        ]
        self.assertEqual(price_display_stored, "$ {:.2f}".format(value2))

        obj.price = Undefined
        price_display_stored = obj.read(("price_display_stored",))[0][
            "price_display_stored"
        ]
        self.assertEqual(price_display_stored, "--")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import logging
from collections import namedtuple, Mapping
from xoutil.string import cut_prefix
from xoutil.objects import import_object

from odoo import fields, api, models
from odoo.fields import Char
from odoo.release import version_info as ODOO_VERSION_INFO


logger = logging.getLogger(__name__)


if ODOO_VERSION_INFO < (12,):
    api_create_signature = api.model
else:
    api_create_signature = api.model_create_multi

__all__ = ["Enumeration"]

Member = namedtuple("Member", ("name", "value"))


class _EnumeratedField(object):
    # I'm keeping this class just in case someone has been testing for it.
    pass


class Enumeration(Char, _EnumeratedField):
    """Create an enumeration field.

    The `enumclass` argument must be either an `enumeration class` or a fully
    qualified name of an enumeration class.  If either way return a callable,
    it must accept a single argument (the model) and return the enumeration
    class.

    The actual enumeration used per model is stored in the ``Enumclass``
    attribute.

    The column in the DB will be of type CHAR and the values are the name of
    attribute in the enumeration class.

    Enumeration classes are required to:

    - have an attribute ``__members__`` with a `collections.Mapping`:class:
      from names to values.

    - have an attribute for each name in ``__member__`` with the same value as
      in the mapping.

    Those rules imply::

       >>> all(getattr(enumclass, name) is value
       ...     for name, value in enumclass.__members__.items())

    These requirements are compatible with Python 3.4's `enum.Enum`:class:
    However, we don't require that values are instances of the enumeration
    class.

    .. rubric:: Automatic selection field

    .. warning:: Known issues with this features.

       Up to 0.66.0 this feature is not yet stable.  The injected field is not
       visible to ``api.depends`` and other components of Odoo framework.

       However, the method `get_selection_field`:meth: seems to work well.

    Enumeration fields cannot be put in views because their values can be
    arbitrary Python objects which are not easily transferred to/from the web
    client.

    You can use the parameter 'selection_field_name' to automatically inject a
    computed Selection field that allows to represent this Enumeration in the
    web client.  This automatic is by default not stored in the DB.  Its
    values are the names of the members of the enumeration class.

    You may pass a function 'compute_member_string' which will be called to
    get the user facing text of each member.  The function will be passed both
    the name and value of the member.  By default we return the member's name.

    Additional keyword arguments starting with the prefix ``selection_field_``
    are used to construct the Selection field.  For instance,
    'selection_field_string' sets the argument 'string'.  If not provided, we
    set 'string', and 'help' to the same values of the Enumeration field.

    .. versionchanged:: 0.36.0 A new generalized enumeration field.  Maintains
       DB compatibility (does not need migrations), but do require changes in
       the code.

    .. versionchanged:: 0.47.0 Add keyword parameter 'force_char_column'.

    .. versionchanged:: 0.61.0 The DB column type is always a CHAR.  Ignore
       the parameter 'force_char_column'.

    .. versionchanged:: 0.61.0 Add an automatic way to create a selection
       field via the parameter 'selection_field_name'.

    .. versionchanged:: 0.64.0 The `enumclass` can be a callable.  Enumeration
       is now class instead of a function.  Gain the ``Enumclass`` attribute.

    .. versionchanged:: 0.66.0 The function `compute_member_string` must now
       take three arguments: the model, the name and the value.

    """

    type = "enumeration"

    _slots = {
        "enumclass": None,
        "selection_field_kwargs": None,
        "compute_member_string": None,
        "Enumclass": None,
    }

    def __init__(self, enumclass, *args, **kwargs):
        selection_field_kwargs = kwargs.pop("selection_field_kwargs", {})
        if not selection_field_kwargs:
            SELECTION_FIELD_PREFIX = "selection_field_"
            selection_field_kwargs = {
                cut_prefix(kw, SELECTION_FIELD_PREFIX): value
                for kw, value in kwargs.items()
                if kw.startswith(SELECTION_FIELD_PREFIX)
            }
            compute_member_string = kwargs.pop("compute_member_string", None)
            if compute_member_string:
                selection_field_kwargs["compute_member_string"] = compute_member_string

        super(Enumeration, self).__init__(
            enumclass=enumclass,
            selection_field_kwargs=selection_field_kwargs,
            *args,
            **kwargs,
        )
        self.enumclass = enumclass
        self.selection_field_kwargs = selection_field_kwargs

    def new(self, **kwargs):
        enumclass = kwargs.pop("enumclass", self.enumclass)
        selection_field_kwargs = kwargs.pop(
            "selection_field_kwargs", self.selection_field_kwargs
        )
        return type(self)(
            enumclass=enumclass, selection_field_kwargs=selection_field_kwargs, **kwargs
        )

    def resolve_enumclass(self, model):
        enumclass = import_object(self.enumclass)

        if callable(enumclass) and not hasattr(enumclass, "__members__"):
            enumclass_factory = enumclass
        else:
            enumclass_factory = constant(enumclass)
        return enumclass_factory(model)

    def get_selection_field(
        self, field_name, selection_field_name, compute_member_string=None, **kwargs
    ):
        """Return a computed Selection field to set/get the Enumeration.

        The result is a standard `odoo.fields.Selection`:class: where the
        values are computed from the enumeration class.

        The argument to `compute_member_string`, if not None, should be a
        callable taking three arguments: the model, the name of a member, the
        value of a member; and it must return the display value of the
        selected member.

        The rest of the keyword arguments are passed to the selection
        unchanged.  We pass by default:

        compute
           Internal compute function that along with `inverse` below
           implements everything we've stated above.

        inverse
           Internal inverse function.

        store
           Set to False to avoid duplicating stuff in the DB -- the *real*
           value in the enumeration field.

        .. warning:: The first two arguments are annoying but required.

           The must be kept in sync with the enumeration field's name and the
           selection field's name.

        Comprehensive example:

        .. code-block:: python

           from enum import Enum
           from xoeuf import fields, models
           from odoo import _

           class Bark:
               __doc__ = _("Bark like a dog")

               @classmethod
               def make_sound(cls):
                   print("Woof!")

           class Meow:
               __doc__ = _("Meow like a cat")

               @classmethod
               def make_sound(cls):
                   print("Meow!")

           class enumclass(Enum):
               DOG = Bark
               CAT = Meow

           def get_cls_name(self, name, value):
              "Return the docstring as the name in the selection"
              doc = getattr(value, "__doc__", None)
              if doc:
                 return _(doc)
              else:
                 return name

           class Model(models.Model):
                animal = fields.Enumeration(enumclass)
                animal_sound = enum.get_selection_field(
                    "animal",
                    "animal_sound",
                    get_cls_name,
                    string="Animal sound"
                )

        """

        @api.multi
        @api.depends(field_name)
        def _compute_selection_field(rs):
            enumclass = self.resolve_enumclass(rs)
            for record in rs:
                value = getattr(record, field_name, None)
                if value is not None:
                    member = _get_member_by_value(enumclass, value)
                    setattr(record, selection_field_name, member.name)
                else:
                    setattr(record, selection_field_name, False)

        @api.multi
        def _set_selection(rs):
            enumclass = self.resolve_enumclass(rs)
            for record in rs:
                key = getattr(record, selection_field_name, None)
                if key:
                    member = _get_member_by_name(enumclass, key)
                    setattr(record, field_name, member.value)
                else:
                    setattr(record, field_name, False)

        if not compute_member_string:
            compute_member_string = self._default_compute_member_string
        kwargs.setdefault("store", False)
        kwargs.setdefault("compute", _compute_selection_field)
        kwargs.setdefault("inverse", _set_selection)
        return fields.Selection(
            selection=lambda s: [
                (name, compute_member_string(s, name, value))
                # Don't use `self.Enumclass`: when the selection field is
                # computed the setup_full may not be called yet.
                for name, value in self.resolve_enumclass(s).__members__.items()
            ],
            **kwargs,
        )

    @staticmethod
    def _default_compute_member_string(self, name, value):
        # The 'self' is actually a parameter that (the model); it's not a typo
        # and this function IS intentionally a staticmethod.
        return name

    def get_member_by_value(self, value, record=None):
        return _get_member_by_value(self.Enumclass, value)

    def get_member_by_name(self, name):
        return _get_member_by_name(self.Enumclass, name)

    def setup_full(self, model):
        # Injects a new class in the model MRO, so that we can guarantee
        # the expectation about the methods create, write and search.
        #
        # The alternative would be to implement pre_save and pre_search
        # receivers.  But that would run for ALL models regardless of the
        # presence of Enumeration fields.  Also, we need to rewrite the
        # argument in the 'search' method, and signals aren't designed for
        # such use case (although they work at the moment).
        #
        # We must find EnumerationAdapter in the MRO to avoid injecting it
        # twice when using Enumeration fields in abstract models.
        #
        # We don't need to inject the EnumerationAdapter in related
        # (delegated) copies for this Enumeration fields, because the
        # actual field does the right thing with DB (it does have the
        # EnumerationAdapter injected.)
        if self._setup_done != "full":
            cls = type(model)
            if EnumerationAdapter not in cls.mro() and not self.related:
                cls.__bases__ = (EnumerationAdapter,) + cls.__bases__
                logger.debug(
                    "Setting %s to model %s(%s, %s)", self, model, cls, id(cls)
                )
            self.Enumclass = self.resolve_enumclass(model)
            result = super(Enumeration, self).setup_full(model)
            if not self.compute and not self.related and not model._abstract:
                assert self.name in model.fields_get()
            return result

    def _add_selection_field(self, model):
        if self.selection_field_kwargs:
            cls = type(model)
            selection_field_name = self.selection_field_kwargs.pop("name")
            if selection_field_name not in cls._fields and not self.related:
                selection_field = self.get_selection_field(
                    self.name, selection_field_name, **self.selection_field_kwargs
                )
                logger.debug("Adding %s to model %s", selection_field_name, model)
                model._add_field(selection_field_name, selection_field)

    def convert_to_read(self, value, record, use_name_get=True):
        if value is not None and value is not False:
            return self.get_member_by_value(value).value
        else:
            return value

    def convert_to_write(self, value, record):
        if value is not None and value is not False:
            if value in self.Enumclass.__members__.values():
                member = self.get_member_by_value(value)
                # Our EnumerationAdapter takes care of doing the right
                # thing when writing to the DB, also convert_to_column
                # does.
                return member.value
        return value

    def convert_to_cache(self, value, record, validate=True):
        if value not in self.Enumclass.__members__.values():
            if value is not None and value is not False:
                return self.get_member_by_name(value).value
        return value

    # NOTE: The parameter `values` was introduced in Odoo 11; the
    # parameter validate was introduced in Odoo 12.  We don't use
    # either; but declare them both to work across the three Odoo
    # versions.
    def convert_to_column(self, value, record, values=None, validate=True):
        if value in self.Enumclass.__members__.values():
            return Char.convert_to_column(
                self, self.get_member_by_value(value).name, record
            )
        else:
            return Char.convert_to_column(self, value, record)


class Adapter(object):
    # See the note in setup_full above.

    # Odoo 11 requires all bases to have these attributes.  See
    # odoo/models.py, method _build_model_attributes in lines 515-555 (at the
    # time of writing).
    _table = None
    _sequence = None
    _inherits = _depends = {}
    _sql_constraints = []
    _constraints = []
    _description = None

    # Odoo 12 requires this attribute.  See odoo/models.py, line 528
    _inherit = None


class EnumerationAdapter(Adapter):
    "Adapt the create/write/search method to Enumeration fields."

    @api.model
    @api.returns(*models.BaseModel.search._returns)
    def search(self, args, *pos_args, **kwargs):
        for index, query_part in enumerate(args):
            if not isinstance(query_part, str):
                fieldname, operator, operands = query_part
                field = self._fields.get(fieldname, None)
                if isinstance(field, Enumeration):
                    if operator in (
                        "=",
                        "!=",
                        "ilike",
                        "not ilike",
                        "like",
                        "not like",
                        "=like",
                        "=ilike",
                    ):
                        values = _get_db_value(field, operands)
                    elif operator in ("in", "not in"):
                        values = [_get_db_value(field, o) for o in operands]
                    else:
                        raise TypeError(
                            "Unsupported operator %r for an enumeration field"
                            % operator
                        )
                    args[index] = (fieldname, operator, values)
        return super(EnumerationAdapter, self).search(args, *pos_args, **kwargs)

    @api_create_signature
    @api.returns("self", lambda value: value.id)
    def create(self, values):
        if isinstance(values, Mapping):
            values = {
                fieldname: _get_db_value(self._fields.get(fieldname, None), value)
                for fieldname, value in values.items()
            }
        else:
            values = [
                {
                    fieldname: _get_db_value(self._fields.get(fieldname, None), value)
                    for fieldname, value in vals.items()
                }
                for vals in values
            ]
        return super(EnumerationAdapter, self).create(values)

    @api.multi
    def write(self, values):
        for fieldname, value in dict(values).items():
            field = self._fields.get(fieldname, None)
            values[fieldname] = _get_db_value(field, value)
        return super(EnumerationAdapter, self).write(values)


def _get_db_value(field, value):
    if value is None or value is False:  # and not field.required
        return value
    if isinstance(field, Enumeration):
        member = field.get_member_by_value(value)
        return member.name
    else:
        return value


def _get_member_by_value(enumclass, value):
    """Find the enumclass's member that is equal to `value`."""
    try:
        return next(
            Member(k, v) for k, v in enumclass.__members__.items() if v == value
        )
    except StopIteration:
        raise ValueError("Invalid member %r of enumeration %r" % (value, enumclass))


def _get_member_by_name(enumclass, name):
    """Find the enumclass's member by name"""
    try:
        return Member(name, enumclass.__members__[name])
    except (AttributeError, KeyError):
        raise ValueError("Invalid key %r of enumeration %r" % (name, enumclass))


def constant(value):
    "Create a function (of many args) that always returns a constant `value`."

    def result(*args, **kwargs):
        return value

    return result


if ODOO_VERSION_INFO < (11, 0):

    @api.model
    def _setup_fields(self, partial):
        cls = type(self)
        for field in dict(cls._fields).values():
            if isinstance(field, Enumeration):
                field._add_selection_field(self)
        _super_setup_fields(self, partial)


else:

    @api.model
    def _setup_fields(self):
        cls = type(self)
        for field in dict(cls._fields).values():
            if isinstance(field, Enumeration):
                field._add_selection_field(self)
        _super_setup_fields(self)


_super_setup_fields = models.BaseModel._setup_fields
models.BaseModel._setup_fields = _setup_fields



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import unittest

from hypothesis import strategies, given
from datetime import datetime

from odoo.tests.common import BaseCase

formats = strategies.sampled_from(("%H:%M", "%H:%M:%S", "%H:%M:%S.%f"))


class TestTimeRange(BaseCase):
    # In Odoo 12 BaseCase is required (or the metaclass MetaCase).  But
    # BaseCase overrides the assertRaises expecting subclasses to put an
    # 'env'; which we do not need.
    assertRaises = unittest.TestCase.assertRaises

    @given(strategies.times(), strategies.times())
    def test_timerange(self, t1, t2):
        from xoeuf.fields.timerange.utils import TimeRange

        start = min(t1, t2)
        end = max(t1, t2)
        timerange = TimeRange(start, end)
        self.assertEqual(eval(repr(timerange)), timerange)
        if start != end:
            with self.assertRaises(ValueError):
                TimeRange(end, start)

    @given(strategies.times(), strategies.times(), formats)
    def test_timerange_fmt(self, t1, t2, fmt):
        from xoeuf.fields.timerange.utils import TimeRange

        # Since any of t1 or t2 will be formatted, for the 'end', I have to make
        # sure that start < end.
        t1 = datetime.strptime(t1.strftime(fmt), fmt).time()
        t2 = datetime.strptime(t2.strftime(fmt), fmt).time()
        start = min(t1, t2)
        end = max(t1, t2)
        TimeRange(start, end.strftime(fmt))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoeuf.odoo.tests.common import BaseCase

try:
    from xoeuf.odoo.addons.base.models.ir_model import FIELD_TYPES
except ImportError:
    try:
        from xoeuf.odoo.addons.base.ir.ir_model import FIELD_TYPES
    except ImportError:
        # Odoo 10 is cool
        FIELD_TYPES = None


class TestNewFieldsAreRegistered(BaseCase):
    def test_getting_the_field_property(self):
        if FIELD_TYPES is not None:
            self.assertIn(("python-property", "python-property"), FIELD_TYPES)

    def test_getting_the_field_timedelta(self):
        if FIELD_TYPES is not None:
            self.assertIn(("timedelta", "timedelta"), FIELD_TYPES)

    def test_getting_the_field_timerange(self):
        if FIELD_TYPES is not None:
            self.assertIn(("timerange", "timerange"), FIELD_TYPES)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# TODO: @manu, At least in Odoo 10, there is 'openerp.sentrylog'
try:
    from odoo.sentrylog import *  # noqa: reexport
except ImportError:
    from ._sentrylog import *  # noqa: reexport



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Xœuf ORM extensions for Open Object (OpenERP) models.

All the attributes of this module must be functions that can be integrated as
`BaseModel` methods or operators.

Operators name has the format ``operator__<name>`` and are converted to
``__<name>__``.

"""
from xoeuf.osv.writers import ORMWriter as get_writer  # noqa

orm_writer = get_writer


from xoeuf.osv.writers import ORMCreator as get_creator  # noqa

orm_creator = get_creator


def get_treeview_action(self, **options):
    """Return the tree view action for `self`.

    If the `self` is a singleton, return the view type 'form' so that it's
    easier to see the full object.

    :rtype: An action `dict` you can return to the web client.

    `options` is used to override **almost any** key in the returned dict.  An
    interesting option is `target`.

    """
    _, _, context = self.env.args
    ids = self.ids
    if not ids or len(ids) > 1:
        vtype = "list"
        active_id = None
    else:
        vtype = "form"
        active_id = ids[0]
    result = {
        "type": "ir.actions.act_window",
        "res_model": self._name,
        "src_model": False,
        "view_id": False,
        "view_type": vtype,
        "view_mode": "list,form",
        "views": [(False, "list"), (False, "form")],
        "target": "current",
        "context": context,
    }
    if ids:
        result["domain"] = [("id", "in", tuple(ids))]
    if active_id:
        result["active_id"] = active_id
    result.update(options)
    return result


def get_ref(self, xmlid, depth=1):
    """Get the XMLID in the caller's addon.

    If `xmlid` is missing the addon, then will try to guess it from the
    `calling addon <xoeuf.modules.get_caller_addon>`:func:.

    `depth` is passed directly to `~xoeuf.modules.get_caller_addon`:func:.

    """
    if "." in xmlid:
        return self.env.ref(xmlid)
    else:
        from xoeuf.modules import get_caller_addon

        addon = get_caller_addon(depth=depth)
        return self.env.ref(addon + "." + xmlid)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# __init__
# ---------------------------------------------------------------------
# Copyright (c) 2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2017-08-01
from . import models  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Xœuf are basic services for OpenERP and Open Object.

The name is composed by:
  * x: The starting letter for almost all Merchise projects.
  * oe: Open and ERP initials.
  * œuf: Is "egg" in french.

"""
from . import modules  # noqa; bootstrap 'xoeuf.odoo'
from . import signals  # noqa; bootstrap signals
from .osv import orm  # bootstrap 'orm' (injects _RELATED in XMLs 'eval')

from xoeuf.odoo import SUPERUSER_ID  # noqa
from xoeuf.odoo.release import version_info as ODOO_VERSION_INFO  # noqa

MAJOR_ODOO_VERSION = ODOO_VERSION_INFO[0]  # noqa

# Bootstrap fields; otherwise they won't appear in the FIELD_TYPES in
# ir_model.py
from . import fields  # noqa

from ._version import get_versions

del orm

__version__ = get_versions()["version"]
del get_versions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from odoo.fields import Html as Base

from lxml import html
from lxml.etree import XMLSyntaxError

from xoeuf.utils import hybridmethod


class Html(Base):
    """An HTML field.

    This is an extension to Odoo's fields.Html to add a couple of methods.

    """

    @hybridmethod
    def extract_text(self_or_cls, record_or_value, raises=True):
        """Extract plain text from an HTML field.

        If given value cannot be parsed and `raises` is True, raise a
        ValueError; if `raises` is False return None.

        This method can be called as classmethod, or as instance method.  In
        the first case, the argument should a string with the contents of the
        HTML.  In the second case, the argument should be a singleton
        recordset from which we extract the field's value.

        .. versionadded:: 0.74.0

        """
        value = self_or_cls._get_html_value(record_or_value)
        if value:
            try:
                texts = html.fromstring(value).xpath("//text()")
            except XMLSyntaxError:
                if raises:
                    raise
                else:
                    return None
            else:
                if texts:
                    return " ".join(texts)
        else:
            return ""

    @hybridmethod
    def is_plain_text_empty(self_or_cls, record_or_value, raises=True):
        """Return True if the plain text of HTML value is empty.

        If `raises` is False and the value is not valid (i.e
        `Html.extract_text`:method: returns None), return None.  If `raises`
        is True, raise a ValueError.

        This method can be called as classmethod, or as instance method.  In
        the first case, the argument should a string with the contents of the
        HTML.  In the second case, the argument should be a singleton
        recordset from which we extract the field's value.

        .. versionadded:: 0.74.0

        """
        res = self_or_cls.extract_text(record_or_value, raises=raises)
        if res is not None:
            return not res.strip()
        else:
            return None

    @hybridmethod
    def _get_html_value(self_or_cls, record_or_value):
        if isinstance(self_or_cls, type) and issubclass(self_or_cls, Base):
            return record_or_value
        else:
            return record_or_value[self_or_cls.name]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from odoo.fields import Monetary as Base


class Monetary(Base):
    """A monetary field.

    This is the same as Odoo's original `odoo.fields.Monetary`:class:, with an
    additional attribute:

    :param concrete: If set to True record-set will return the value as a
                     `concrete monetary value <xoutil.dim.currencies>`:mod:.
                     Which means you can only operate it with commensurable
                     monetary values.

    """

    _slots = {"concrete": False}

    # Putting the concrete quantity in the cache (i.e implementing the
    # convert_to_cache) may lead to unwanted type errors (TypeError:
    # incomparable quantities: 0.0::{EUR}/{} and 0) when rounding the value in
    # the cache.  So the best approach is just to override the __get__ of the
    # descriptor.
    def __get__(self, instance, owner):
        from xoutil.dim.currencies import currency as Currency

        result = super(Monetary, self).__get__(instance, owner)
        if instance and self.concrete:
            result *= Currency(instance[self.currency_field].name)
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from collections import Mapping
from enum import IntEnum, Enum
from xoeuf import api, models, fields, MAJOR_ODOO_VERSION


class COLORS(IntEnum):
    Blue = 0
    Red = 1
    Green = 2


class CARS(Enum):
    FORD = object()
    CHEV = object()


class Pax(object):
    single = 1
    double = 2
    __members__ = {"single": single, "double": double}


class WorkType(int):
    pass


class WORK_TYPE(object):
    easy = WorkType(0)
    hard = WorkType(1)
    __members__ = {"easy": easy, "hard": hard}


class Mixin(models.AbstractModel):
    _name = "test_enumeration.mixin"
    color = fields.Enumeration(COLORS, default=COLORS.Red)
    color_name = color.get_selection_field("color", "color_name")
    # color_rgb = fields.Char(compute="_compute_color_rgb")
    # color_rgb2 = fields.Char(compute="_compute_color_rgb", store=True)

    # def _compute_color_rgb(self):
    #     for record in self:
    #         if record.color_name == "Red":
    #             record.color_rgb = "#f00"
    #         elif record.color_name == "Blue":
    #             record.color_rgb = "#00f"
    #         elif record.color_name == "Green":
    #             record.color_rgb = "#0f0"
    #         else:
    #             record.color_rgb = "#fff"


class Model(models.Model):
    _name = "test.enum.model"
    _inherit = ["test_enumeration.mixin"]

    car = fields.Enumeration(CARS)
    pax = fields.Enumeration(Pax)

    def _get_enumclass(self):
        if self._name == "test.enum.model_delegated":

            class Dynamic(Enum):
                name = "Dynamic"

        else:

            class Dynamic(Enum):
                name = "Static"

        return Dynamic

    dynamic_enum = fields.Enumeration(_get_enumclass)

    # def _compute_color_rgb(self):
    #     for record in self:
    #         record.color_rgb = "#000"


if MAJOR_ODOO_VERSION < 12:
    api_create_signature = api.model
else:
    api_create_signature = api.model_create_multi


class DelegatedModel(models.Model):
    _name = "test.enum.model_delegated"
    _inherits = {"test.enum.model": "model_id"}

    model_id = fields.Many2one("test.enum.model")

    @api_create_signature
    def create(self, values):
        if isinstance(values, Mapping):
            values = self._pre_create_model_id(values)
        else:
            values = [self._pre_create_model_id(vals) for vals in values]
        return super(DelegatedModel, self).create(values)

    @api.model
    def _pre_create_model_id(self, values):
        if "model_id" not in values:
            values["model_id"] = self.env["test.enum.model"].create({}).id
        return values



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#


def TimeSpan(start_date_field, end_date_field):
    """Create a time span `property <xoeuf.fields.Property>`:class:.

    A time span `property <xoeuf.fields.Property>`:class: is stored in the DB
    as two separate Date fields.  It's never actually stored in the DB as a
    single value.  You cannot search for this property.

    :param start_date_field: The name of the field which holds the start date
                             of the time span.

    :param end_date_field: The name of the field which holds the end date.

    .. note:: None of those fields are created automatically.  You must create
              them.

    Example::

       class SomeModel(Model):
            some_start_date = fields.Date()
            some_end_date = fields.Date()
            some = fields.TimeSpan('some_start_date', 'some_end_date')

    """
    from xoeuf.fields import Property
    from xoutil.future.datetime import TimeSpan

    @Property
    def result(self):
        return TimeSpan(getattr(self, start_date_field), getattr(self, end_date_field))

    @result.setter
    def result(self, value):
        setattr(self, start_date_field, value.start_date)
        setattr(self, end_date_field, value.end_date)

    @result.deleter
    def result(self):
        # This may fail if any of these fields is required.
        setattr(self, start_date_field, None)
        setattr(self, end_date_field, None)

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from odoo.fields import Many2one


class One2one(Many2one):
    """A classical one-2-one field.

    This is basically syntax sugar over a `~odoo.fields.Many2one`:class: with:

    - `ondelete` set to ``'cascade'`` by default;

    - a SQL uniqueness constraint in the model.

    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("ondelete", "cascade")
        super(One2one, self).__init__(*args, **kwargs)

    def setup_full(self, model):
        res = super(One2one, self).setup_full(model)
        if self.related is None:
            constraint_name = "unique_{model}_{target}".format(
                model=model._name.replace(".", "_"),
                target=self.comodel_name.replace(".", "_"),
            )
            constraints = type(model)._sql_constraints
            exists = any(True for name, _, _ in constraints if name == constraint_name)
            if not exists:
                constraints.append(
                    (
                        constraint_name,
                        "unique ({field})".format(field=self.name),
                        "One2one field {field!r} must be unique".format(
                            field=self.name
                        ),
                    )
                )
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

{
    "name": "test_field_html_extensions",
    "version": "1.0",
    "depends": ["base"],
    "auto_install": False,
    "installable": True,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""User interface stuff.

"""


class _const(dict):
    def __call__(self):
        return self


# Return this to close a window
CLOSE_WINDOW = _const({"type": "ir.actions.act_window_close"})


# Return this to reload the interface
RELOAD_UI = _const({"type": "ir.actions.client", "tag": "reload"})


# The do nothing action.
DO_NOTHING = _const({})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import unittest

from xoutil.future.codecs import safe_decode

from xoeuf import MAJOR_ODOO_VERSION
from xoeuf.odoo.tests.common import TransactionCase, at_install, post_install
from xoeuf.signals import (
    mock_replace,
    post_create,
    pre_create,
    write_wrapper,
    pre_fields_view_get,
    no_signals,
)
from xoeuf.odoo.addons.test_signals.models import (
    post_save_receiver,
    post_save_receiver_all_models,
    pre_save_receiver,
    wrap_nothing,
    pre_fvg_receiver,
)


@at_install(False)
@post_install(True)
class TestXoeufSignals(TransactionCase):
    def setUp(self):
        super(TestXoeufSignals, self).setUp()
        self.Model = self.env["test_signals.signaling_model"]
        self._ready_before = self.env.registry.ready
        self.env.registry.ready = True

    def tearDown(self):
        self.env.registry.ready = self._ready_before

    def test_post_create(self):
        with mock_replace(
            post_create, post_save_receiver
        ) as mock_post_create, mock_replace(
            pre_create, post_save_receiver
        ) as mock_pre_create:
            self.Model.create(dict(name="My name"))
            self.assertTrue(mock_post_create.called)
            self.assertFalse(mock_pre_create.called)

    def test_pre_create(self):
        with mock_replace(
            post_create, pre_save_receiver
        ) as mock_post_create, mock_replace(
            pre_create, pre_save_receiver
        ) as mock_pre_create:
            self.Model.create(dict(name="My name"))
            self.assertFalse(mock_post_create.called)
            self.assertTrue(mock_pre_create.called)

    def test_writer_wrap(self):
        def wrapper(sender, signal, *args, **kwargs):
            yield

        who = self.Model.create(dict(name="My name"))
        with mock_replace(write_wrapper, wrap_nothing, side_effect=wrapper) as mock:
            self.assertFalse(mock.called)
            who.write(dict(name="My new name"))
            self.assertTrue(mock.called)
        with mock_replace(write_wrapper, wrap_nothing, side_effect=wrapper) as mock:
            who.name = "My new other name"
            self.assertTrue(mock.called)

    def test_not_called_for_partners(self):
        with mock_replace(post_create, post_save_receiver) as mock, mock_replace(
            post_create, post_save_receiver_all_models
        ) as mock_all:
            partner = self.env["res.partner"].create(dict(name="Contact Info"))
            partner.email = "a@b.c"
            self.assertFalse(mock.called)
            self.assertTrue(mock_all.called)

    def test_fvg_in_abstract_models(self):
        who = self.Model.create(dict(name="My name"))
        with mock_replace(pre_fields_view_get, pre_fvg_receiver) as mock:
            result = who.fields_view_get()
            self.assertTrue(mock.called)
            self.assertIn("fgv-is-present", safe_decode(result["arch"], "utf-8"))

    def test_no_signals(self):
        # Test mocks outside and inside
        with mock_replace(pre_create, pre_save_receiver) as mock:
            with no_signals(pre_create):
                self.Model.create(dict(name="My name"))
            self.Model.create(dict(name="My name"))
            self.assertEqual(mock.call_count, 1)

        with no_signals(pre_create):
            with mock_replace(pre_create, pre_save_receiver) as mock:
                self.Model.create(dict(name="My name"))
                self.assertFalse(mock.called)

    @unittest.skipIf(MAJOR_ODOO_VERSION < 12, "create accepts a list from Odoo 12+")
    def test_model_create_multi(self):
        with mock_replace(pre_create, pre_save_receiver) as mock:
            result = self.Model.create([dict(name="My name"), dict(name="Second obj")])
            # The mock is called once even though there are two records to
            # create.
            self.assertEqual(mock.call_count, 1)
            self.assertEqual(len(result), 2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from psycopg2 import IntegrityError
from xoeuf.odoo.tests.common import TransactionCase


class TestO2O(TransactionCase):
    def test_one2one_injects_restriction(self):
        C = self.env["test.one2one.c"]
        c = C.create({"name": "I am a C"})
        a = c.a_id
        with self.assertRaises(IntegrityError):
            C.create({"name": "I am trying to steal the A", "a_id": a.id})

    def test_one2one_baseline(self):
        B = self.env["test.one2one.b"]
        b = B.create({"name": "I am a B"})
        a = b.a_id
        B.create({"name": "I can steal the A", "a_id": a.id})

    def test_one2one_cascade(self):
        C = self.env["test.one2one.c"]
        c = C.create({"name": "I am a C"})
        a = c.a_id
        a.unlink()
        self.assertFalse(c.exists())

    def test_one2one_related(self):
        D = self.env["test.one2one.d"]
        d = D.create({"name": "I am a D"})
        a = d.a_id
        with self.assertRaises(IntegrityError):
            D.create({"name": "I am trying to steal the A", "a_id": a.id})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoeuf import fields, models, api

MIXIN_NAME = "typed.reference.example.mixin"


class ExampleMixin(models.AbstractModel):
    _name = MIXIN_NAME

    test = fields.Char(default="Hello")
    partner_id = fields.Many2one("res.partner")
    test2 = fields.Char(default="Hello")
    name = fields.Char(related="partner_id.name", readonly=False)
    comment = fields.Text(related="partner_id.comment", readonly=False)
    ref = fields.Char(related="partner_id.ref", store=True, readonly=False)
    street = fields.Char(related="partner_id.street", store=True, readonly=False)
    phone = fields.Char(related="partner_id.phone", readonly=False)
    mobile = fields.Char(related="partner_id.mobile", readonly=False)


class Model1(models.Model):
    _name = "test.model1"
    _inherit = MIXIN_NAME
    _description = "Model 1"


class Model2(models.Model):
    _name = "test.model2"
    _inherit = MIXIN_NAME


class SubModel2(models.Model):
    _name = "test.sub.model2"
    _inherit = "test.model2"


class TestModel(models.Model):
    _name = "test.model"

    typed_ref = fields.TypedReference(mixin=MIXIN_NAME, delegate=True)
    filtered_typed_ref = fields.TypedReference(
        mixin=MIXIN_NAME, selection=[("test.model2", "model2")]
    )
    # define to set store=True
    test2 = fields.Char(compute="_compute_test2", store=True)
    # name = fields.Char(compute="_compute_name", store=True)
    ref = fields.Char(compute="_compute_ref", store=True)

    @api.one
    @api.depends("typed_ref", "typed_ref.test2")
    def _compute_test2(self):
        if self.typed_ref:
            self.test2 = self.typed_ref.test2

    # @api.one
    # @api.depends("typed_ref", "typed_ref.partner_id.name")
    # def _compute_name(self):
    #     if self.typed_ref:
    #         self.name = self.typed_ref.name

    @api.one
    @api.depends("typed_ref", "typed_ref.ref")
    def _compute_ref(self):
        if self.typed_ref:
            self.ref = self.typed_ref.ref



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from itertools import product
import logging

from hypothesis import strategies as s, given, settings
from hypothesis.stateful import RuleBasedStateMachine, rule, Bundle
from hypothesis.stateful import run_state_machine_as_test

from xoeuf.osv import expression as expr
from xoeuf.osv import ql
from xoeuf.odoo.osv import expression as odoo_expr
from xoeuf.osv.expression import Domain, DomainTree

from odoo.tests.common import TransactionCase, BaseCase

names = s.text(alphabet="abdefgh", min_size=1, max_size=5)
operators = s.sampled_from(["=", "!=", "<", ">", "<>"])
all_operators = s.sampled_from(
    ["=", "!=", "<", ">", "<>", "like", "ilike", "not like", "not ilike"]
)
ages = s.integers(min_value=0, max_value=120)
sensible_values = s.integers(min_value=-10, max_value=10)

# Logical connectors with the amount of terms it connects.  Notice that ''
# takes two arguments because it's the same as '&'.
connectors = s.sampled_from([("", 2), ("&", 2), ("!", 1), ("|", 2)])

logger = logging.getLogger(__name__)


@s.composite
def terms(draw, fields=None, values=None):
    if not values:
        values = sensible_values
    f = fields or names
    name = str(draw(f))
    operator = draw(operators)
    value = draw(values)
    return (name, operator, value)


@s.composite
def domains(draw, fields=None, min_size=1, max_size=10):
    leaves = draw(s.lists(terms(fields), min_size=min_size, max_size=max_size))
    result = []
    while leaves:
        connector, many = draw(connectors)
        if many <= len(leaves):
            if connector:
                result.append(connector)
            result.extend(leaves[:many])
            leaves = leaves[many:]
        elif len(leaves) == 1:
            result.append(leaves.pop(0))
    return Domain(result)


class TestDomain(BaseCase):
    @given(domains())
    def test_first_normal_form_idempotency(self, domain):
        self.assertEqual(
            domain.first_normal_form, domain.first_normal_form.first_normal_form
        )

    @given(domains())
    def test_second_normal_form_idempotency(self, domain):
        self.assertEqual(
            domain.second_normal_form, domain.second_normal_form.second_normal_form
        )

    @given(domains())
    def test_versions_equivalency(self, domain):
        # Any domain must be equivalent in all its versions.
        all_versions = [
            domain,
            domain.simplified,
            domain.first_normal_form,
            domain.second_normal_form,
        ]
        for version1, version2 in product(all_versions, all_versions):
            self.assertEqual(version1, version2, msg="%r != %r" % (version1, version2))

    @settings(deadline=1000)
    @given(domains())
    def test_simplified(self, domain):
        # A simplified domain never start with '&' operator.
        self.assertNotEqual(
            domain.simplified[0],
            expr.AND_OPERATOR,
            msg="%r start with &" % domain.simplified,
        )
        # The simplified domain must be the shortest.
        all_versions = [domain, domain.first_normal_form, domain.second_normal_form]
        for version in all_versions:
            self.assertLessEqual(
                len(domain.simplified),
                len(version),
                msg="len(%r) > len(%r)" % (domain.simplified, version),
            )
        self.assertTrue(
            all(
                odoo_expr.is_leaf(term) or odoo_expr.is_operator(term)
                for term in domain.simplified
            )
        )

    @given(domains())
    def test_simplified_idempotency(self, domain):
        self.assertEqual(domain.simplified, domain.simplified.simplified)

    @given(s.lists(domains(), min_size=2, max_size=10))
    def test_expr_replacement(self, domains):
        expr.AND(domains)
        odoo_expr.AND(domains)

    def test_eq(self):
        A = Domain([("field_x", "=", 1)])
        # A == A
        self.assertTrue(A == A)

        # ~A == ~A
        self.assertTrue(~A == ~A)

        # (x = 1)  == ! (x <> 1)
        A1 = Domain(["!", ("field_x", "<>", 1)])
        self.assertTrue(A == A1)

    def test_implies(self):
        # Basic implications.
        # -------------------

        A = Domain([("field_x", "=", 1)])
        B = Domain([("field_y", "!=", False)])

        # A => A
        self.assertTrue(A.implies(A))

        # A => A | B
        self.assertTrue(A.implies(A | B))

        # A & B => A
        self.assertTrue((A & B).implies(A))

        # A & B => A | B
        self.assertTrue((A & B).implies(A | B))

        self.assertEqual(Domain.AND(list(A), B), (A & B))
        self.assertEqual(expr.AND([list(A), B]), (A & B))

        # Inter terms implications.
        # -------------------------

        # (x = 1)  => (x = 1)
        A = Domain([("field_x", "=", 1)])
        A1 = Domain([("field_x", "=", 1)])
        self.assertTrue(A.implies(A1))

        # (x != 1)  => (x <> 1)
        A = Domain([("field_x", "!=", 1)])
        A1 = Domain([("field_x", "<>", 1)])
        self.assertTrue(A.implies(A1))

        # (x > 1) => (x > 0)
        A = Domain([("field_x", ">", 10)])
        A1 = Domain([("field_x", ">", 1)])
        self.assertTrue(A.implies(A1))

        # (x in (1,2,3)) => (x in (1,2,3,4))
        A = Domain([("field_x", "in", (1, 2, 3))])
        A1 = Domain([("field_x", "in", (1, 2, 3, 4))])
        self.assertTrue(A.implies(A1))
        #

        # Compound implications.
        # ----------------------
        # A => B & C iff (A => B) & (A => C)
        A = Domain([("field_x", "in", (1, 4))])
        B = Domain([("field_x", "in", (1, 2, 4))])
        C = Domain([("field_x", "in", (1, 3, 4))])
        self.assertTrue(A.implies(B))
        self.assertTrue(A.implies(C))
        self.assertTrue(A.implies(B & C))

        # More complex implications
        # -------------------------

        x = Domain(
            [
                "|",
                ("field_y", "!=", False),
                ("field_x", "=", "value"),
                ("field_z", "in", (1,)),
            ]
        )
        y = Domain(
            [
                "|",
                ("field_y", "!=", False),
                ("field_x", "=", "value"),
                ("field_z", "in", (1, 2, 3)),
                "|",
                ("field_x", "=", "value"),
                ("field_y", "!=", False),
            ]
        )
        assert x.implies(y)

        x = Domain(
            [
                ("field_y", "!=", False),
                ("field_z", "in", (1,)),
                ("field_z", "in", (1, 2)),
                "|",
                ("field_y", "!=", False),
                "|",
                ("field_z", "in", (1,)),
                ("field_z", "in", (2,)),
            ]
        )
        y = Domain([("field_y", "!=", False), ("field_z", "in", (1, 2, 3))])
        assert x.implies(y)

    def test_walk(self):
        y = Domain(
            [
                "|",
                ("a", "!=", False),
                ("b", "=", "value"),
                ("c", "in", (1, 2, 3)),
                "|",
                "|",
                "!",
                ("d", "=", "value"),
                ("f", "!=", False),
                ("g", "=", "h"),
            ]
        )
        DomainTree(y.second_normal_form)
        # expected = [
        #     ('TERM', ('c', 'in', (1, 2, 3))),
        #     ('TERM', ('g', '=', 'h')),
        #     ('TERM', ('d', '!=', 'value')),
        #     ('TERM', ('f', '!=', False)),
        #     ('OPERATOR', '|'),
        #     ('OPERATOR', '|'),
        #     ('TERM', ('a', '!=', False)),
        #     ('TERM', ('b', '=', 'value')),
        #     ('OPERATOR', '|'),
        #     ('OPERATOR', '&'),
        #     ('OPERATOR', '&')
        # ]
        # self.assertEqual(expected, list(tree.walk()))

    def test_get_filter_ast_simple_one_term_with_in(self):
        self.assertEqual(
            DomainTree(Domain([("state", "in", [1, 2])]))._get_filter_ast(),
            ql.Expression(
                ql.Lambda(
                    ql.make_arguments("this"),
                    expr._constructor_in("this", "state", [1, 2]),
                )
            ),
        )
        Domain([("state", "in", [1, 2])]).asfilter()

    def test_empty_domain(self):
        self.assertTrue(Domain([]).asfilter()(0))


def get_model_domain_machine(this):
    Model = this.env["test_domain.model"]

    class ModelDomainMachine(RuleBasedStateMachine):
        objects = Bundle("objects")

        @rule(target=objects, name=names, age=ages)
        def create_object(self, name, age):
            logger.info("Creating object name: %s, age: %s", name, age)
            return Model.create({"name": name, "age": age})

        @rule(target=objects, name=names, age=ages, parent=objects)
        def create_child_object(self, name, age, parent):
            return Model.create({"name": name, "age": age, "parent_id": parent.id})

        @rule(age=ages, op=operators)
        def find_by_age(self, age, op):
            query = Domain([("age", op, age)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(
            age=ages,
            op=operators,
            path=s.lists(
                s.sampled_from(["parent_id", "children_ids"]), min_size=1, max_size=4
            ),
        )
        def find_by_parent_age(self, age, op, path):
            attr = ".".join(path) + ".age"
            query = Domain([(attr, op, age)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(ages=s.lists(ages))
        def find_by_ages(self, ages):
            query = Domain([("age", "in", ages)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(
            ages=s.lists(ages),
            path=s.lists(
                s.sampled_from(["parent_id", "children_ids"]), min_size=1, max_size=4
            ),
        )
        def find_by_parent_ages(self, ages, path):
            attr = ".".join(path) + ".age"
            query = Domain([(attr, "in", ages)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(ages=s.lists(ages))
        def find_by_not_ages(self, ages):
            query = Domain([("age", "not in", ages)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(
            ages=s.lists(ages),
            path=s.lists(
                s.sampled_from(["parent_id", "children_ids"]), min_size=1, max_size=4
            ),
        )
        def find_by_path_not_ages(self, ages, path):
            attr = ".".join(path) + ".age"
            query = Domain([(attr, "not in", ages)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(domain=domains(fields=s.just("age")))
        def find_by_arbitrary_domain(self, domain):
            res = Model.search(domain)
            logger.info("Check filter/domain: %s; count: %s", domain, len(res))
            this.assertEqualRecordset(res.filtered(domain.asfilter()), res)

        @rule(name=names, op=all_operators)
        def find_by_name(self, name, op):
            query = Domain([("name", op, name)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(
            name=names,
            op=all_operators,
            path=s.lists(
                s.sampled_from(["parent_id", "children_ids"]), min_size=1, max_size=4
            ),
        )
        def find_by_path_name(self, name, op, path):
            attr = ".".join(path) + ".name"
            query = Domain([(attr, op, name)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(names=s.lists(names))
        def find_by_names(self, names):
            query = Domain([("name", "in", names)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(
            names=s.lists(names),
            path=s.lists(
                s.sampled_from(["parent_id", "children_ids"]), min_size=1, max_size=4
            ),
        )
        def find_by_path_names(self, names, path):
            attr = ".".join(path) + ".name"
            query = Domain([(attr, "in", names)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(names=s.lists(names))
        def find_by_not_names(self, names):
            query = Domain([("name", "not in", names)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

        @rule(
            names=s.lists(names),
            path=s.lists(
                s.sampled_from(["parent_id", "children_ids"]), min_size=1, max_size=4
            ),
        )
        def find_by_path_not_names(self, names, path):
            attr = ".".join(path) + ".name"
            query = Domain([(attr, "not in", names)])
            res = Model.search(query)
            logger.info("Check filter/domain: %s; count: %s", query, len(res))
            this.assertEqualRecordset(res.filtered(query.asfilter()), res)

    return ModelDomainMachine


class TestConsistencyOfFilters(TransactionCase):
    def assertEqualRecordset(self, rs1, rs2):
        ours = rs1 - rs2
        theirs = rs2 - rs1
        self.assertEqual(
            rs1, rs2, msg="ours: {0!r}; theirs: {1!r}".format(ours, theirs)
        )

    def test_consistency_of_domains(self):
        run_state_machine_as_test(get_model_domain_machine(self))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Xœuf services for access Open Object (OpenERP) models.

Implement the network protocols that the Xœuf applications that access OpenERP
databases uses to communicate with its clients.

"""
import pytz
from xoeuf.tools import dt_as_timezone


def savepoint(cr, name=None):
    """A context manager that enters a new savepoint.

    If `name` is not provided a random one is generated.

    .. note:: In Odoo v8, the cursor `cr` has already a ``savepoint`` method.
              This function will fallback to it in this case.  However, the
              argument `name` will be ignored in this case.

    """
    from xoutil.symbols import Unset

    _savepoint = getattr(cr, "savepoint", Unset)
    if _savepoint is Unset:
        from contextlib import contextmanager

        if not name:
            from xoutil.uuid import uuid

            name = uuid(True)

        @contextmanager
        def _savepoint():
            cr.execute('SAVEPOINT "%s"' % name)
            try:
                yield
            except:  # noqa
                cr.execute('ROLLBACK TO SAVEPOINT "%s"' % name)
                raise
            else:
                cr.execute('RELEASE SAVEPOINT "%s"' % name)

    return _savepoint()


def datetime_user_to_server_tz(cr, uid, userdate, tz_name=None, context=None):
    """ Convert date values expressed in user's timezone to
    server-side UTC timestamp.

    :param datetime userdate: datetime in user time zone
    :return: UTC datetime for server-side use
    """
    from xoeuf import api

    utc = pytz.UTC
    if userdate.tzinfo:
        return utc.normalize(userdate)
    if not tz_name:
        env = api.Environment(cr, uid, context or {})
        user = env.user
        if user.tz:
            dt = dt_as_timezone(userdate, user.tz)
        else:
            dt = dt_as_timezone(userdate)
    else:
        dt = dt_as_timezone(userdate, tz_name)
    return utc.normalize(dt)


def datetime_server_to_user_tz(cr, uid, serverdate, tz_name=None, context=None):
    """ Convert date values expressed in server-side UTC timestamp to
    user's timezone.

    :param datetime serverdate: datetime in server-side UTC timestamp.
    :return: datetime on user's timezone
    """
    from xoeuf import api

    dt = dt_as_timezone(serverdate)  # datetime in UTC
    if not tz_name:
        env = api.Environment(cr, uid, context or {})
        user = env.user
        user_tz = pytz.timezone(user.tz) if user.tz else pytz.UTC
    else:
        user_tz = pytz.timezone(tz_name)
    return user_tz.normalize(dt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from odoo.tests.common import BaseCase


class TestXoeufImports(BaseCase):
    def test_xoeuf_imports(self):
        "Xoeuf is importable in Python 2 and 3"
        from xoeuf import fields, models, api  # noqa

        try:
            from xoeuf.fields import Serialized  # noqa
        except ImportError:
            pass
        else:
            assert False, "Serialized should not be exported"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoeuf import fields
from xoeuf.odoo.tests.common import TransactionCase
from .. import MIXIN_NAME


class TestReferences(TransactionCase):
    def test_simple_typed_reference(self):
        obj = self.env["test.model"].create({})
        typed_ref = obj._fields["typed_ref"]
        self.assertTrue(isinstance(typed_ref, fields.TypedReference))
        self.assertEqual(typed_ref.mixin, MIXIN_NAME)

        def typed_ref_assign(value):
            obj.typed_ref = value
            return True

        self.assertTrue(typed_ref_assign(False))
        obj2 = self.env["test.model1"].create({})
        self.assertTrue(typed_ref_assign(obj2))
        self.assertTrue(typed_ref_assign("%s,%d" % (obj2._name, obj2.id)))
        self.assertRaises(ValueError, typed_ref_assign, obj)

    def test_filtered_typed_reference(self):
        obj = self.env["test.model"].create({})

        def typed_ref_assign(value):
            obj.filtered_typed_ref = value
            return True

        obj2 = self.env["test.model1"].create({})
        # This should work, but the odoo validation is missing,
        # I am making a PR to propose it.
        # self.assertRaises(ValueError, typed_ref_assign, '%s,%d' % (obj2._name, obj2.id))
        self.assertRaises(ValueError, typed_ref_assign, obj2)

    def test_delegate_typed_reference(self):
        obj = self.env["test.model"].create({})
        obj2 = self.env["test.model1"].create({"test": "ok"})
        dummy_value = "any value"
        obj.test = dummy_value
        self.assertEqual(obj.test, dummy_value)
        obj.invalidate_cache()
        self.assertFalse(obj.test)
        obj.typed_ref = obj2
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertTrue(obj._fields["typed_ref"].delegate)
        self.assertTrue(obj._fields["test"].compute)
        self.assertEqual(obj.test, obj2.test)
        obj.test = "other value"
        self.assertEqual(obj.test, obj2.test)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj.test, obj2.test)
        obj2.test = "other more value"
        # todo: indirect depends are not working fine.
        # self.assertEqual(obj.test, obj2.test)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj.test, obj2.test)

    def test_related_field_setup(self):
        obj = self.env["test.model"]
        obj2 = self.env["test.model1"]
        # check related field attributes like: string, help, comodel_name, ...
        for attr, _ in obj._fields["partner_id"].related_attrs:
            self.assertEqual(
                getattr(obj._fields["partner_id"], attr),
                getattr(obj2._fields["partner_id"], attr),
            )

    def test_related_field_search(self):
        # computed non store field from a regular field
        self.env["test.model"].search([("test", "=", "any value")])
        # computed non store field from a computed non store field
        # self.env["test.model"].search([('comment', '=', 'any value')])  # no searchable
        # computed non store field from a computed store field
        self.env["test.model"].search([("street", "=", "any value")])
        # computed store field from a regular field
        self.env["test.model"].search([("test2", "=", "any value")])
        # computed store field from a computed store field
        self.env["test.model"].search([("ref", "=", "any value")])
        # computed store field from a computed non store field
        # self.env["test.model"].search([("name", "=", "any value")])

    def test_related_field_right_triggers(self):
        """All `store=False` computed field through a `Reference` are computed correctly.

        This behave is the same with any `Reference` odoo field.

        """
        obj = self.env["test.model1"].create(
            {"partner_id": self.env.user.partner_id.id}
        )
        obj2 = self.env["test.model"].create({"typed_ref": obj.reference_repr})
        dummy_value = "any value"

        # computed non store field from a regular field
        obj.test = dummy_value
        self.assertEqual(obj2.test, dummy_value)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj2.test, dummy_value)

        # computed non store field from a computed non store field
        obj.comment = dummy_value
        self.assertEqual(obj2.comment, dummy_value)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj2.comment, dummy_value)
        obj.partner_id.comment = dummy_value
        self.assertEqual(obj2.comment, dummy_value)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj2.comment, dummy_value)

        # computed non store field from a computed store field
        obj.street = dummy_value
        self.assertEqual(obj2.street, dummy_value)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj2.street, dummy_value)
        obj.partner_id.street = dummy_value
        self.assertEqual(obj2.street, dummy_value)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj2.street, dummy_value)

    def test_related_field_wrong_triggers(self):
        """No `store=True` computed field through a `Reference` is computed correctly.

        The computed `store=True` fields from a `store=False` field through a
        `Reference` cause an` AttributeError` when writing on the source field.

        This behave is the same with any `Reference` odoo field.

        """
        obj = self.env["test.model1"].create(
            {"partner_id": self.env.user.partner_id.id}
        )
        obj2 = self.env["test.model"].create({"typed_ref": obj.reference_repr})
        dummy_value = "any value"

        # computed store field from a regular field
        obj.test2 = dummy_value
        self.assertNotEqual(obj2.test2, dummy_value)  # Is not updated
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertNotEqual(obj2.test2, dummy_value)  # Is not updated

        # computed store field from a computed store field
        obj.ref = dummy_value
        self.assertNotEqual(obj2.ref, dummy_value)  # Is not updated
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertNotEqual(obj2.ref, dummy_value)  # Is not updated

        # computed store field from a computed non store field
        # with self.assertRaises(AttributeError):
        #     obj.name = dummy_value
        # with self.assertRaises(AttributeError):
        #     obj.partner_id.name = dummy_value

    def test_related_field_inverse(self):
        obj = self.env["test.model1"].create(
            {"partner_id": self.env.user.partner_id.id}
        )
        obj2 = self.env["test.model"].create({"typed_ref": obj.reference_repr})
        dummy_value = "any value"

        # computed non store field from a regular field
        obj2.test = dummy_value
        self.assertEqual(obj.test, dummy_value)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj.test, dummy_value)

        # computed non store field from a computed non store field
        obj2.comment = dummy_value
        self.assertEqual(obj.comment, dummy_value)
        self.assertEqual(obj.partner_id.comment, dummy_value)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj.comment, dummy_value)
        self.assertEqual(obj.partner_id.comment, dummy_value)

        # computed non store field from a computed store field
        obj2.street = dummy_value
        self.assertEqual(obj.street, dummy_value)
        self.assertEqual(obj.partner_id.street, dummy_value)
        obj.invalidate_cache()
        obj2.invalidate_cache()
        self.assertEqual(obj.street, dummy_value)
        self.assertEqual(obj.partner_id.street, dummy_value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
try:
    from ._version import get_version

    VERSION = get_version()
except:  # noqa
    from ._version import get_versions

    VERSION = get_versions()["version"]

RELEASE_TAG = ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Transparently import models.

Example usage (in xoeuf' shell)::

    $ xoeuf shell
    >>> from xoeuf.models.proxy import AccountAccount

    # We need a 'self' (and possible cr, uid) in the context of any method
    # call to model.
    >>> AccountAccount.search([])   # doctest: +ELLIPSIS
    Traceback (...)
    ...
    AttributeError: search


    $ xoeuf shell -d somedb
    >>> from xoeuf.models.proxy import ResUsers as User

    >>> User.search([], limit=1)  # doctest: +ELLIPSIS
    res.users(...)


"""
import re
from os.path import splitext
from types import ModuleType
from . import _proxy

import warnings

warnings.warn("xoeuf.model.proxy is deprecated and will be removed before 1.0")


class ModelImporter(object):
    @classmethod
    def hook(cls, path):
        if path == __path__[0]:
            return cls
        else:
            raise ImportError

    @classmethod
    def find_module(cls, name, path=None):
        base, _sep, name = name.rpartition(str("."))
        if base == __name__:
            return cls(name, base)
        else:
            msg = 'We can not load "%s" outside "%s".'
            raise ImportError(msg % (name, __name__))

    def load_module(self, fullname):
        """Returns the loaded module or raises an exception."""
        import sys

        self._check(fullname)
        res = sys.modules.get(fullname, None)
        if res is None:
            res = ModelProxy(fullname.rsplit(str("."), 1)[-1])
            res.__path__ = __path__
            sys.modules[fullname] = res
        return res

    def __init__(self, name, base):
        self.__name = name
        self.__base = base

    @property
    def fullname(self):
        sep = str(".")
        return sep.join((self.__base, self.__name))

    def _check(self, fullname, asserting=False):
        "Check `fullname` and raise an error if not correct."
        local = self.fullname
        if local == fullname:
            return fullname
        elif asserting:
            return False
        else:
            msg = 'Using loader for "%s" to manage "%s"!'
            raise ImportError(msg % (self.name, fullname))


class ModelProxy(ModuleType, _proxy.ModelProxy):
    def __init__(self, name):
        ModuleType.__init__(self, name)
        _proxy.ModelProxy.__init__(self, name)


__path__ = [splitext(__file__)[0]]


import sys

sys.path_hooks.append(ModelImporter.hook)
del sys, ModelImporter, splitext, re



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoeuf import fields

from odoo.tests.common import TransactionCase


class TestLocalizedDt(TransactionCase):
    def setUp(self):
        super().setUp()
        self.valid_html = self.env["test_field_html_extensions.model"].create(
            {"html": "<p>This is ok</p>"}
        )
        self.empty_html = self.env["test_field_html_extensions.model"].create(
            {"html": ""}
        )
        self.empty_html_with_tags = self.env["test_field_html_extensions.model"].create(
            {"html": "<p> &nbsp;</p>"}
        )

    def test_valid_html_non_empty_record_signature(self):
        field = self.valid_html._fields["html"]
        self.assertEqual(field.extract_text(self.valid_html), "This is ok")
        self.assertFalse(field.is_plain_text_empty(self.valid_html))

    def test_valid_html_non_empty_recordless_signature(self):
        self.assertEqual(fields.Html.extract_text(self.valid_html.html), "This is ok")
        self.assertFalse(fields.Html.is_plain_text_empty(self.valid_html.html))

    def test_valid_empty_html_without_tags_record_signature(self):
        field = self.valid_html._fields["html"]
        self.assertEquals(field.extract_text(self.empty_html), "")
        self.assertTrue(field.is_plain_text_empty(self.empty_html))

    def test_valid_empty_html_without_tags_recordless_signature(self):
        self.assertEquals(fields.Html.extract_text(self.empty_html.html), "")
        self.assertTrue(fields.Html.is_plain_text_empty(self.empty_html.html))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""A new object model.


"""
from . import proxy  # noqa: ensure the hook is registered.
from .base import *  # noqa: reexport

from xoeuf.odoo.models import *  # noqa: reexport



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from . import test_view_model  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import hypothesis
from hypothesis import strategies as s
from xoutil.dim.currencies import currency as Currency

from xoeuf.odoo.tests.common import TransactionCase

finite_floats = s.floats(
    min_value=-(2 ** 32), max_value=2 ** 32, allow_nan=False, allow_infinity=False
)


class TestMonetary(TransactionCase):
    def assertValueInCurrency(self, value, currency):
        """Assert value is expressed in units of `currency`."""
        unit = Currency(currency.name)
        try:
            # If can sum the currency's unit to the value, it's expressed in
            # the same currency unit.
            value + unit
        except TypeError:
            raise AssertionError(
                "'%s' is not in terms of currency '%s'" % (value, currency.name)
            )

    @hypothesis.given(finite_floats)
    @hypothesis.settings(max_examples=5)
    def test_monetary_concrete(self, value):
        Line = self.env["test.monetary.concrete"]
        EUR = self.env.ref("base.EUR")
        obj = Line.create({"value": value, "currency_id": EUR.id})
        self.assertValueInCurrency(obj.value, EUR)

    @hypothesis.given(finite_floats)
    @hypothesis.settings(max_examples=5)
    def test_monetary_related(self, value):
        Line = self.env["test.monetary.related"]
        company = self.env.ref("base.main_company")
        EUR = company.currency_id = self.env.ref("base.EUR")
        obj = Line.create({"value": value, "company_id": company.id})
        self.assertValueInCurrency(obj.value, EUR)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoeuf import fields, models


class Mixin(models.AbstractModel):
    _name = "test_localized_dt.mixin"

    tzone = fields.Char(default="America/Havana")
    dt = fields.Datetime()
    dt_at_tzone = fields.LocalizedDatetime("dt", "tzone")


class ModelA(models.Model):
    _name = "test_localizated_dt.model"
    _inherit = ["test_localized_dt.mixin"]


class ModelB(models.Model):
    _name = "test_localized_dt.inherited"
    _inherits = {models.get_modelname(ModelA): "a_id"}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
try:
    from xoutil.objects import crossmethod  # TODO: migrate
except ImportError:

    class crossmethod(object):
        """Decorate a function as static or instance level.

        Example:

          >>> class Mule(object):
          ...     @crossmethod
          ...     def print_args(*args):
          ...         print(args)

          # Call it as a staticmethod
          >>> Mule.print_args()
          ()

          # Call it as an instance
          >>> Mule().print_args()   # doctest: +ELLIPSIS
          (<...Mule object at ...>,)

        .. note:: This is different from `hybridmethod`:func:.  Hybrid method
                  always receive the implicit argument (either `cls` or
                  `self`).

        """

        def __init__(self, func):
            self.func = func

        def __get__(self, instance, owner):
            if instance is None:
                return self.func
            else:
                return self.func.__get__(instance, owner)


try:
    from xoutil.objects import hybridmethod  # TODO: migrate
except ImportError:
    # Code taken from SQLAlchemy
    #
    # Copyright (C) 2005-2016 the SQLAlchemy authors and contributors
    #
    # SQLAlchem is released under the MIT License:
    # http://www.opensource.org/licenses/mit-license.php

    class hybridmethod:
        """Decorate a function as cls- or instance- level."""

        def __init__(self, func):
            self.func = func

        def __get__(self, instance, owner):
            if instance is None:
                return self.func.__get__(owner, owner.__class__)
            else:
                return self.func.__get__(instance, owner)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
try:
    from xoutil.future.datetime import TimeRange  # TODO: migrate
except ImportError:
    from datetime import datetime, time

    class TimeField(object):
        """A descriptor for a `datetime.time`:class:.

        You may set either `datetime.time`:class: values or strings with the
        'HH:MM[:SS[.µs]]' format.

        `name` is the attribute name.  This is the key in the instance's
        __dict__.

        If `nullable` is True (the default).  You may also set, None or False
        (Odoo).

        .. note:: tzinfo is not supported.

        """

        def __init__(self, name, nullable=True):
            self.name = name
            self.nullable = nullable

        def __get__(self, instance, owner):
            try:
                return instance.__dict__[self.name]
            except KeyError:
                raise AttributeError(self.name)

        def __set__(self, instance, value):
            if value in (None, False):
                if not self.nullable:
                    raise ValueError(
                        "Setting None to a non nullable attribute %r" % self.name
                    )
            elif isinstance(value, str):
                if "." in value:
                    fmt = "%H:%M:%S.%f"
                elif len(value) > len("99:99"):
                    fmt = "%H:%M:%S"
                else:
                    fmt = "%H:%M"
                value = datetime.strptime(value, fmt).time()
            elif not isinstance(value, time):
                raise TypeError(
                    "Either time or str expected. Got %r" % type(value).__name__
                )
            instance.__dict__[self.name] = value

        def __delete__(self, instance):
            del instance.__dict__[self.name]

    class TimeRange(object):
        """A continuous `datetime.time`:class: range.

        Both `start` and `end` are inclusive.

        .. note:: So far we don't support unbounded ranges.

        """

        start = TimeField("start", False)
        end = TimeField("end", False)

        def __init__(self, start, end):
            self.start = start
            self.end = end
            if self.start > self.end:
                raise ValueError("start must be earlier than end.")

        def __bool__(self):
            return True

        __nonzero__ = __bool__

        def __contains__(self, value):
            return self.start <= value <= self.end

        def __repr__(self):
            start, end = self.start, self.end
            return "TimeRange(%r, %r)" % (
                self._format_time(start),
                self._format_time(end),
            )

        @staticmethod
        def _format_time(time):
            from xoutil.string import cut_suffix

            return cut_suffix(time.isoformat(), ":00")

        def __eq__(self, other):
            if isinstance(other, TimeRange):
                return self.start == other.start and self.end == other.end
            else:
                return NotImplemented

        # TODO: All others operators like __and__, __lshift__, and the empty
        # timerange, which is only needed to make __and__ closed.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import logging
import numbers
import pickle

from xoeuf import models, fields, api
from xoeuf.odoo.tools.safe_eval import safe_eval

logger = logging.getLogger(__name__)
del logging


class _Undefined:
    _instance = None

    def __new__(cls):
        if cls._instance is not None:
            return cls._instance
        else:
            res = cls._instance = super().__new__(cls)
            return res


Undefined = _Undefined


class PriceMixin(models.AbstractModel):
    _name = "test.property.pricemixin"

    # We'll store the pickled value in the DB; the true value is in the price
    # Property and a user-friendly repr in the price field.
    _db_price = fields.Binary()

    @fields.Property
    def price(self):
        if self._db_price:
            return pickle.loads(self._db_price)
        else:
            return Undefined

    @price.setter
    def price(self, value):
        if isinstance(value, numbers.Number):
            self._db_price = pickle.dumps(value)
        else:
            self._db_price = None

    price_display = fields.Char(compute="_compute_price", inverse="_set_price")
    price_display_stored = fields.Char(
        compute="_compute_price", inverse="_set_price", store=True
    )

    @api.one
    @api.depends("price")
    def _compute_price(self):
        if self.price is Undefined:
            self.price_display = "--"
            self.price_display_stored = "--"
        else:
            self.price_display = "$ {:.2f}".format(self.price)
            self.price_display_stored = "$ {:.2f}".format(self.price)

    def _set_price(self):
        user_price = self.price_display.strip()
        if user_price and user_price != "--" and user_price[0] in "$01234567890.":
            if user_price[0] == "$":
                user_price = user_price[1:].strip()
            try:
                if "." in user_price:
                    self.price = float(user_price)
                else:
                    self.price = int(user_price)
            except ValueError:
                self.price = Undefined


class ValueMixin(models.AbstractModel):
    _name = "test.property.valuemixin"

    value = fields.Char()

    @fields.Property
    def inverted(self):
        return "".join(reversed(self.value)) if self.value else self.value

    @inverted.setter
    def inverted(self, value):
        self.value = "".join(reversed(value)) if value else value

    @inverted.deleter
    def inverted(self):
        self.value = None

    @inverted.onsetup
    def inverted(field, model):
        type(model).inverted_setup = True

    @fields.Property(memoize=True)
    def memoized_object(self):
        return object()

    @memoized_object.setter
    def memoized_object(self, value):
        pass

    @memoized_object.deleter
    def memoized_object(self):
        pass

    @fields.Property
    def new_object(self):
        return object()


class ObjectMixin(models.AbstractModel):
    _name = "test.property.object"

    thing = fields.Char(
        # This stores the representation (result of repr) of the value.  You
        # may use `result` to read/update this field.  If you use Odoo's
        # write/create, or set this field directly you MUST ensure to pass the
        # result of `repr()`.  The `write/create` methods below ensure the
        # writing to 'result' works.
    )

    @api.multi
    def write(self, values):
        # Makes writing to 'result' work.  If you pass both 'value' and
        # 'result', 'result' wins.
        if "result" in values:
            values["thing"] = repr(values.pop("result"))
        return super(ObjectMixin, self).write(values)

    @api.model
    def create(self, values):
        # Makes writing to 'result' work.  If you pass both 'value' and
        # 'result', 'result' wins.
        if "result" in values:
            values["thing"] = repr(values.pop("result"))
        return super(ObjectMixin, self).create(values)

    def _get_result(self):
        return safe_eval(self.thing) if self.thing else self.thing

    def _set_result(self, value):
        self.thing = repr(value)

    def _del_result(self):
        self.thing = None

    result = fields.Property(
        getter=_get_result, setter=_set_result, deleter=_del_result
    )
    del _get_result, _set_result, _del_result


class Value(models.Model):
    _name = "test.property.value"
    _inherit = [ValueMixin._name, ObjectMixin._name, PriceMixin._name]


class IdentifiedValue(models.Model):
    _name = "test.property.inherits"
    _inherits = {"test.property.value": "value_id"}

    name = fields.Char(required=True)
    value_id = fields.Many2one("test.property.value", required=True, ondelete="cascade")

    _sql_constrains = [
        ("unique_name", "UNIQUE(name)", "Names cannot be repeated"),
        ("single_value", "UNIQUE(value_id)", "A value can only appear once"),
    ]


class DelegatedValue(models.Model):
    _name = "test.property.delegated"
    name = fields.Char(required=True)
    value_id = fields.Many2one(
        "test.property.value", required=True, ondelete="cascade", delegate=True
    )

    _sql_constrains = [
        ("unique_name", "UNIQUE(name)", "Names cannot be repeated"),
        ("single_value", "UNIQUE(value_id)", "A value can only appear once"),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Profiling tools.

For development only purposes.

"""
from xoutil.decorator.meta import decorator


try:
    from line_profiler import LineProfiler

    profiler = LineProfiler()

    @decorator
    def profile(func, model=True, stream=None):
        """Decorate `func` to activate the line profiler.

        :param func: The function to decorate

        :param model: Set to False if this function is not contained in an
                      Odoo model class definition.

        :param stream:  The stream into which print the collected stats.  The
                        default is whatever LineProfiler has as default.

        """
        from xoeuf import api

        if model:
            f = getattr(func, "_orig", func)
            wraps = api.mimic
        else:
            f = func
            wraps = lambda f: lambda x: x  # noqa

        profiler.add_function(f)

        @wraps(func)
        def inner(*args, **kwargs):
            profiler.enable_by_count()
            try:
                return f(*args, **kwargs)
            finally:
                profiler.disable_by_count()
                profiler.print_stats(stream)

        return inner


except ImportError:
    import warnings

    profiler = None

    @decorator
    def profile(func, model=None, stream=None):
        """Decorate `func` to activate the line profiler.

        """
        warnings.warn(
            "Trying to the profiler without the line_profiler package.  "
            "Are you trying to use the profiler in production?! "
            "Don't do that EVER AGAIN!!"
        )
        return func


del decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import unittest
import datetime
import pytz

from hypothesis import strategies, given, example
from hypothesis.extra.pytz import timezones
from xoeuf.tools import localtime_as_remotetime, normalize_datetime

from odoo.tests.common import BaseCase


class TestLocalAsRemote(BaseCase):
    @unittest.skip("Known failure")
    @given(
        strategies.datetimes(
            min_value=normalize_datetime("1912-04-14"),
            max_value=normalize_datetime("2100-12-31"),
        ),
        timezones(),
        timezones(),
    )
    @example(
        _dt=datetime.datetime(1946, 1, 1, 0, 0),
        from_tz=pytz.timezone("Europe/Kaliningrad"),
        to_tz=pytz.UTC,
    )
    def test_local_remote_time(self, _dt, from_tz, to_tz):
        _dt = normalize_datetime(_dt)
        self.assertEqual(
            _dt,
            localtime_as_remotetime(
                localtime_as_remotetime(_dt, from_tz, to_tz), to_tz, from_tz
            ),
        )
        self.assertEqual(
            _dt,
            localtime_as_remotetime(
                localtime_as_remotetime(_dt, from_tz, to_tz, ignore_dst=True),
                to_tz,
                from_tz,
                ignore_dst=True,
            ),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Secures a DB by:

- Changing all passwords

- [TODO] Changing all accounting related data by a random mapping.

"""
from . import Command


class Secure(Command):
    """The secure command.

    """

    @classmethod
    def get_arg_parser(cls):
        from xoutil.crypto import PASS_LEVEL_NAME_MAPPING
        from xoutil.crypto import DEFAULT_PASS_PHRASE_LEVEL

        def path(extensions=None):
            """A type-builder for file arguments."""
            from xoutil.future.types import is_collection
            from os.path import abspath, isfile, splitext

            if extensions and not is_collection(extensions):
                extensions = (extensions,)
            acceptable = lambda ext: not extensions or ext in extensions

            def inner(value):
                res = abspath(value)
                name, extension = splitext(value)
                if not isfile(res) or not acceptable(extension):
                    raise TypeError("Invalid filename %r" % res)
                return res

            return inner

        res = getattr(cls, "_arg_parser", None)
        if not res:
            from argparse import ArgumentParser

            res = ArgumentParser()
            cls._arg_parser = res
            res.add_argument(
                "-c",
                "--config",
                dest="conf",
                required=False,
                default=None,
                type=path(),
                help="A configuration file.  This could be "
                "either a Python file, like that required by "
                "Gunicorn deployments, or a INI-like "
                'like the standard ".openerp-serverrc".',
            )
            res.add_argument(
                "-l",
                "--level",
                dest="security_level",
                choices=PASS_LEVEL_NAME_MAPPING.keys(),
                default=DEFAULT_PASS_PHRASE_LEVEL,
                help="The security level for passwords.  If "
                '"basic" the password will the same as the '
                'user\'s login.  "random" means a truly random '
                "password.",
            )
            res.add_argument(
                "-d",
                "--database",
                dest="database",
                type=cls.database_factory,
                required=True,
            )
            loggroup = res.add_argument_group("Logging")
            loggroup.add_argument(
                "--log-level",
                choices=("debug", "warning", "info", "error"),
                default="warning",
                help="How much to log",
            )
        return res

    @classmethod
    def database_factory(cls, database):
        from xoeuf.odoo import api, SUPERUSER_ID
        from xoeuf.odoo.modules.registry import Registry

        db = Registry(database)
        env = api.Environment(db.cursor(), SUPERUSER_ID, {})
        return env["res.users"]

    def run(self, args=None):
        from xoeuf.security import reset_all_passwords

        parser = self.get_arg_parser()
        options = parser.parse_args(args)
        conffile = options.conf
        level = options.security_level
        if conffile:
            self.read_conffile(conffile)
        self = options.database
        with self.env.cr:
            reset_all_passwords(self, security_level=level)

    def read_conffile(self, filename):
        import os

        ext = os.path.splitext(filename)[-1]
        if ext == ".py":
            self.load_config_from_script(filename)
        else:
            self.load_config_from_inifile(filename)

    @staticmethod
    def load_config_from_script(filename):
        cfg = {
            "__builtins__": __builtins__,
            "__name__": "__config__",
            "__file__": filename,
            "__doc__": None,
            "__package__": None,
        }
        try:
            with open(filename, "rb") as fh:
                return exec(compile(fh.read(), filename, "exec"), cfg, cfg)
        except Exception:
            import traceback
            import sys

            print("Failed to read config file: %s" % filename)
            traceback.print_exc()
            sys.exit(1)

    @staticmethod
    def load_config_from_inifile(filename):
        from xoeuf.odoo.tools import config

        config.rcfile = filename
        config.load()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoeuf import fields, models
from xoeuf.odoo.addons.test_view_model import TEST_MODEL_NAME


class Model(models.Model):
    _inherit = TEST_MODEL_NAME

    new_value = fields.Float()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from datetime import timedelta
from odoo.fields import Float


class TimeDelta(Float):
    """A timedelta field.

    At the DB the field is represented as an Integer (total_seconds).  In
    Python code we get a timedelta object.

    """

    type = "timedelta"  # Q: Do we need to change the type?

    @staticmethod
    def ensure_timedelta(value):
        """ If `value` is not an timedelta value is assume as float that
        represent the quantity of seconds.

        :param value:
            * timedelta or float are expected.

        :raises TypeError:
            * If value is no an expected type.
        """
        if not isinstance(value, timedelta):
            return timedelta(seconds=value or 0)
        return value

    def convert_to_cache(self, value, *args, **kwargs):
        value = self._convert_to("cache", value, *args, **kwargs)
        return self.ensure_timedelta(value)

    def convert_to_read(self, value, *args, **kwargs):
        return self._convert_to("read", value, *args, **kwargs)

    def convert_to_write(self, value, *args, **kwargs):
        return self._convert_to("write", value, *args, **kwargs)

    def convert_to_column(self, value, *args, **kwargs):
        return self._convert_to("column", value, *args, **kwargs)

    def _convert_to(self, method, value, *args, **kwargs):
        if isinstance(value, timedelta):
            # Ensure pass value as float val.
            value = value.total_seconds()
        return getattr(super(TimeDelta, self), "convert_to_%s" % method)(
            value, *args, **kwargs
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoeuf import models, fields


class TimeSpanValue(models.Model):
    _name = "test.timespan.value"

    start_date = fields.Date()
    end_date = fields.Date()
    period = fields.TimeSpan("start_date", "end_date")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from odoo.tests.common import BaseCase


class TestModules(BaseCase):
    def test_get_object_module(self):
        from .. import Foo
        from xoeuf.modules import get_object_module

        self.assertEqual(get_object_module(Foo), Foo.module)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (
    division as _py3_division,
    print_function as _py3_print,
    absolute_import as _py3_abs_import,
)

import os
import versioneer

from setuptools import setup, find_packages


# Import the version from the release module
project_name = "xoeuf"
_current_dir = os.path.dirname(os.path.abspath(__file__))


def safe_read(*paths):
    fname = os.path.join(_current_dir, *paths)
    try:
        with open(fname, "r") as fh:
            return fh.read()
    except (IOError, OSError):
        return ""


setup(
    name=project_name,
    version=versioneer.get_version(),
    description="Basic utilities for OpenERP Open Object Services",
    long_description=safe_read("docs", "readme.txt"),
    cmdclass=versioneer.get_cmdclass(),
    # Get more strings from
    # http://pypi.python.org/pypi?:action=list_classifiers
    classifiers=[
        "Programming Language :: Python",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",  # noqa
        "Development Status :: 4 - Beta",
    ],
    keywords="openerp open-object server library".split(),
    author="Merchise Autrement [~º/~]",
    author_email="",
    url="http://www.merchise.org/",
    license="GPL",
    packages=find_packages(exclude=["ez_setup", "examples", "tests"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "pytz",
        "decorator>=4.0,<4.2",
        'xoutil>=1.9.0,<2.0; python_version < "3.4"',
        'xoutil>=1.9.0,<2.2; python_version >= "3.4"',
        'ipython<6; python_version < "3"',
        'ipython; python_version >= "3"',
        "raven>=5.8.0",
        "raven-sanitize-openerp",
        'enum34; python_version < "3.4"',
        "celery>=4.1.0,<5",
        'typing;python_version<"3.5"',
    ],
    extra_requires={
        "odoo": [
            'odoo==10.0; python_version < "3.0"',
            'odoo>=11.0,<13.0; python_version >= "3.5"',
        ],
        "test": ["hypothesis>=3.7.0,<4"],
    },
    python_requires=">=3.6",
    entry_points="""
      [console_scripts]
      xoeuf = xoeuf.cli.server:server
      xoeuf_mailgate = xoeuf.cli.mailgate:main

      [xoeuf.addons]
      test_localized_dt = xoeuf.tests.test_localized_dt
      """,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoeuf import models, fields


class A(models.Model):
    _name = "test.one2one.a"
    name = fields.Char()


class B(models.Model):
    _name = "test.one2one.b"
    _inherits = {"test.one2one.a": "a_id"}


class C(models.Model):
    _name = "test.one2one.c"
    _inherits = {"test.one2one.a": "a_id"}

    a_id = fields.One2one("test.one2one.a")


class D(models.Model):
    _name = "test.one2one.d"
    _inherits = {"test.one2one.c": "c_id"}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from odoo.tests.common import TransactionCase


class TestFromActiveIds(TransactionCase):
    # I don't really need to create records, since we're not reading anything
    # from the the DB

    def test_self_takes_from_active_ids(self):
        Model = self.env["xoeuf.tests.test_api.model"]
        res = Model.with_context(
            active_model=Model._name, active_ids=(1, 2, 3, 4)
        ).return_self_ids()
        self.assertEqual(res, [1, 2, 3, 4])

    def test_not_leaked_context(self):
        Model = self.env["xoeuf.tests.test_api.model"]
        this = Model.with_context(active_model=Model._name, active_ids=(1, 2, 3, 4))
        res, res2 = this.return_ids_and_call_method((5, 6), "return_self_ids")
        self.assertEqual(res, [1, 2, 3, 4])
        self.assertEqual(res2, [5, 6])

    def test_leaked_context(self):
        Model = self.env["xoeuf.tests.test_api.model"]
        this = Model.with_context(active_model=Model._name, active_ids=(1, 2, 3, 4))
        res, res2 = this.leaked_return_ids_and_call_method((5, 6), "return_self_ids")
        self.assertEqual(res, [1, 2, 3, 4])
        self.assertEqual(res2, res)

    def test_onupdate(self):
        user = self.env.user
        text_field = "text_field"

        # No onupdate method should be called
        user.text_field = text_field
        self.assertEqual(user.text_field, text_field)

        # ``update_text_field`` must be called
        user.name = "john doe"
        self.assertNotEqual(user.text_field, text_field)
        self.assertEqual(user.text_field, user.get_text_field())

        # ``update_text_field`` must be called
        user = user.create({"login": "new_user", "partner_id": user.partner_id.id})
        self.assertNotEqual(user.text_field, text_field)
        self.assertEqual(user.text_field, user.get_text_field())

        # ``update_text_field`` must be called replacing `text_field` creation value
        user = user.create(
            {
                "login": "new_user2",
                "partner_id": user.partner_id.id,
                "text_field": "text_field",
            }
        )
        self.assertNotEqual(user.text_field, text_field)
        self.assertEqual(user.text_field, user.get_text_field())

        # No onupdate method should be called
        text_field = user.text_field
        user.login = "jane doe"
        self.assertEqual(user.text_field, text_field)

        # ``update_text_field`` must be called
        partner = user.partner_id
        partner.name = "Admin"
        self.assertNotEqual(user.text_field, text_field)
        self.assertEqual(user.text_field, user.get_text_field())

        # No onupdate method should be called
        text_field = user.text_field
        partner.res = "Administrador"
        self.assertEqual(user.text_field, text_field)

    def test_onupdate_inherited_via_mixin(self):
        user = self.env.user
        obj = self.env["text.onupdate.big.model"].create({"user_id": user.id})
        user.name = name = "Johny"
        self.assertEqual(obj.name, "Updated: {name}".format(name=name))

    def test_onupdate_validate_fields(self):
        user = self.env.user
        # Just check we don't raise an exception.
        user._validate_fields("unknown_field_name_" + str(id(self)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoeuf import fields, models


class TestTimeRange(models.Model):
    _name = "test.time.range"

    time_value = fields.Float()
    range_value = fields.TimeRange(
        time_field="time_value",
        selection=[
            ("morning", "Morning", "6:00", "11:59"),
            ("noon", "Noon", "12:00", "13:59"),
            ("afternoon", "Afternoon", "14:00", "18:59"),
        ],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xoeuf import fields
from xoeuf.odoo.tests.common import TransactionCase


class TestProperty(TransactionCase):
    def setUp(self):
        super(TestProperty, self).setUp()
        Value = self.Value = self.env["test.property.value"]
        self.obj = Value.create({"value": "0123456789"})
        for val in "abcdefg":
            Value.create({"value": val})

    def test_setup(self):
        self.assertTrue(getattr(type(self.Value), "inverted_setup", False))

    def test_getter(self):
        obj = self.obj
        self.assertEqual(obj.inverted, "9876543210")

    def test_setter(self):
        obj = self.obj
        obj.inverted = "abcd"
        self.assertEqual(obj.value, "dcba")

    def test_deleter(self):
        obj = self.obj
        del obj.inverted
        self.assertIn(obj.value, (None, False))  # Odoo may xform None to False

    def test_nocreate(self):
        with self.assertRaises(AssertionError):
            # I would like that calling create would raise an exception, but
            # for now it just does not do anything.
            obj = self.Value.create({"inverted": "abcd"})
            self.assertEqual(obj.value, "dcba")

    def test_nowrite(self):
        with self.assertRaises(AssertionError):
            # I would like that calling search would raise an exception, but
            # for now it just does not do anything.
            self.obj.write({"inverted": "abcd"})
            self.assertEqual(self.obj.value, "dcba")

    # Don't perform the test of `search()` because it does not raise an
    # exception, but log an error that may fool our test scripts.

    def test_non_memoized_object(self):
        self.assertIsNot(self.obj.new_object, self.obj.new_object)

    def test_memoized_object(self):
        previous = self.obj.memoized_object
        self.assertIs(self.obj.memoized_object, previous)
        self.obj.memoized_object = sentinel = object()
        self.assertIs(self.obj.memoized_object, sentinel)
        self.assertIs(self.obj.memoized_object, self.obj.memoized_object)
        del self.obj.memoized_object
        self.assertIsNot(self.obj.memoized_object, sentinel)
        self.assertIsNot(self.obj.memoized_object, previous)

    def test_invalidated_memoized_object(self):
        previous = self.obj.memoized_object
        self.obj.invalidate_cache()
        self.assertIsNot(previous, self.obj.memoized_object)


class TestInheritedValue(TransactionCase):
    def setUp(self):
        super(TestInheritedValue, self).setUp()
        Value = self.Value = self.env["test.property.value"]
        Identifier = self.Identifier = self.env["test.property.inherits"]
        for i in range(10):
            val = Value.create({"thing": repr({i: i})})
            iden = Identifier.create({"name": "v%d" % i, "value_id": val.id})
            self.assertEqual(val.thing, iden.thing)

    def test_result_write(self):
        i = self.Identifier.search([], limit=1)
        v = {"a": [1, (1, 2), "a"]}
        i.write({"result": v})
        self.assertEqual(i.thing, repr(v))

    def test_isinstance(self):
        TValue = type(self.env["test.property.value"])
        self.assertIsInstance(TValue.result, fields.Property)
        self.assertNotIsInstance(TValue.thing, fields.Property)

    def test_issubclass(self):
        TValue = type(self.env["test.property.value"])
        self.assertIsSubclass(type(TValue.result), fields.Property)
        self.assertNotIsSubclass(type(TValue.thing), fields.Property)

    def assertIsSubclass(self, subclass, superclass, msg=None):
        assert issubclass(
            subclass, superclass
        ), "{subclass} is not a subclass or {superclass}".format(
            subclass=subclass, superclass=superclass
        )

    def assertNotIsSubclass(self, subclass, superclass, msg=None):
        assert not issubclass(
            subclass, superclass
        ), "{subclass} is a subclass or {superclass}".format(
            subclass=subclass, superclass=superclass
        )
